#pragma once
#ifndef PARSODUS_LRTABLES_GENERATOR_H_YW3GIUNH
#define PARSODUS_LRTABLES_GENERATOR_H_YW3GIUNH

#include "Parsodus/grammar.h"
#include "Parsodus/lrtables/table.h"

#include <queue>

namespace pds {
namespace lr {

const std::string EXTENDED_START = "^";
const std::string EOF_PLACEHOLDER = "$";

/**
 * Base class for LR (and derivative) table generators (such as SLR and LALR)
 * Parametrized on the type of item to be used in the configuration sets
 */
template <typename Item>
class Generator {
    public:
        /**
         * Constructor
         *
         * @param start The start symbol for the grammar
         * @param g The grammar to translate
         */
        Generator(const std::string& start, const Grammar& g);

        /**
         * Generate an LRTable based on given grammar
         *
         * @returns An LR (or derivative) table for the grammar given at construction
         */
        LRTable generate();

    protected:
        /**
         * Indicate whether this specific algorithm needs to have the First and Follow sets generated
         */
        virtual bool needsFollowSet() = 0;

        /**
         * Build the starting item to build all item sets from
         *
         * @param startrule The constructed extended starting rule
         * @param eof The token used as end of file
         */
        virtual Item initial_item(Rule startrule, std::string eof) = 0;


        std::set<std::string> first(std::string s);
        std::set<std::string> follow(std::string s);

    private:
        /**
         * Build the `First` set
         */
        void buildFirst();

        /**
         * Build the `Follow` set
         */
        void buildFollow();

        /**
         * Compute the closure of an item set
         */
        std::set<Item> closure(const std::set<Item>& its);

        Grammar m_gram;
        Rule m_startrule;
        std::map<std::string, std::set<std::string>> m_first;
        std::map<std::string, std::set<std::string>> m_follow;
};

template <typename Item>
Generator<Item>::Generator(const std::string& start, const Grammar& g) : m_gram(g), m_startrule(Rule{EXTENDED_START, {start}}), m_first(), m_follow() {
    m_gram.terminals.insert(EOF_PLACEHOLDER); //End of file
    m_gram.variables.insert(EXTENDED_START); //modified start rule
    m_gram.rules[EXTENDED_START].insert(m_startrule);

    if (needsFollowSet()) {
        buildFirst();
        buildFollow();
    }
}

template <typename Item>
LRTable Generator<Item>::generate() {
    //TODO
    LRTable table;
    std::vector<std::set<Item>> itemsets;
    itemsets.emplace_back(closure({initial_item(m_startrule, EOF_PLACEHOLDER)}));

    std::queue<Item> itemqueue;
    itemqueue.push(itemsets[0]);
    while (!itemqueue.empty()) {
        std::set<Item> cur = std::move(itemqueue.front());
        itemqueue.pop();
        
        for (std::pair<std::string, std::set<Item>> succ : successors(cur)) {
            //Add new itemset or merge
        }
    }

    return table;
}

template <typename Item>
void Generator<Item>::buildFirst() {
    for (std::string term : m_gram.terminals) {
        m_first[term].insert(term);
    }


    bool changes = true;

    auto update = [&changes, this](std::string head, auto& s) {
        for (std::string elem : s) {
            if (!m_first[head].count(elem)) {
                changes = true;
                m_first[head].insert(s.begin(), s.end());
                return;
            }
        }
    };

    while (changes) {
        changes = false;

        for (const auto& p : m_gram.rules) {
            const std::string& head = p.first;
            const std::set<Rule>& rules = p.second;
            for (const Rule& rule : rules) {
                if (rule.tail.size() == 0) {
                    update(head, {""});
                }

                bool ended = false;
                for (const std::string& replacement : rule.tail) {
                    if (m_first[replacement].count("")) {
                        std::set<std::string> tmp = m_first[replacement];
                        tmp.erase("");
                        update(head, tmp);
                    } else {
                        update(head, m_first[replacement]);
                        ended = true;
                        break;
                    }
                }
                if (!ended) {
                    update(head, {""});
                }
            }
        }
    }
}

template <typename Item>
void Generator<Item>::buildFollow() {
    //EOF follow the added start rule.
    m_follow[EXTENDED_START].insert(EOF_PLACEHOLDER);

    bool changes = true;

    auto update = [&changes, this](std::string head, auto s) {
        s.erase("");
        for (std::string elem : s) {
            if (!m_follow[head].count(elem)) {
                changes = true;
                m_follow[head].insert(s.begin(), s.end());
                return;
            }
        }
    };

    while (changes) {
        changes = false;
        for (const auto& p : m_gram.rules) {
            std::string& head = p.first;
            for (const auto& rule : p.second) {
                for (std::size_t i = 0; i < rule.tail.size(); i++) {
                    if (!m_gram.terminals.count(rule.tail[i])) {
                        if (i == rule.tail.size() - 1 || m_first[rule.tail[i + 1]].count("")) {
                            update(rule.tail[i], m_follow[head]);
                        }
                        if (i < rule.tail.size() - 1) {
                            update(rule.tail[i], m_first[rule.tail[i + 1]]);
                        }
                    }
                }
            }
        }
    }
}


} /* lr */
} /* pdf */

#endif /* PARSODUS_LRTABLES_GENERATOR_H_YW3GIUNH */
